@@ -4,7 +4,7 @@ import inject
 from mfcloud.events import EventBus
 
 from twisted.internet import reactor, defer
-from twisted.internet.defer import inlineCallbacks
+from twisted.internet.defer import inlineCallbacks, AlreadyCalledError, CancelledError
 
 from autobahn.twisted.websocket import WebSocketServerFactory
 from autobahn.twisted.websocket import WebSocketClientFactory
@@ -19,15 +19,14 @@ class ApiError(Exception):
     pass
 
 
-
 class ApiRpcServer(object):
-
     redis = inject.attr(txredisapi.Connection)
     eb = inject.attr(EventBus)
 
     def __init__(self):
         self.tasks = {}
         self.ticket_map = {}
+        self.tasks_running = {}
 
         self.eb.on('log-*', self.on_log)
 
@@ -38,16 +37,50 @@ class ApiRpcServer(object):
     def task_completed(self, result, ticket_id):
         if ticket_id in self.ticket_map:
             self.ticket_map[ticket_id].send_event('task.success.%s' % ticket_id, result)
+            del self.tasks_running[ticket_id]
+            del self.ticket_map[ticket_id]
 
     def task_failed(self, error, ticket_id):
         if ticket_id in self.ticket_map:
-            self.ticket_map[ticket_id].send_event('task.failure.%s' % ticket_id, str(error))
+            if isinstance(error, CancelledError):
+                s = 'Terminated.'
+            else:
+                s = str(error)
+
+            self.ticket_map[ticket_id].send_event('task.failure.%s' % ticket_id, s)
+
+            del self.tasks_running[ticket_id]
+            del self.ticket_map[ticket_id]
 
     def task_progress(self, data, ticket_id):
         if ticket_id in self.ticket_map:
             log.msg('Progress: %s' % data)
             self.ticket_map[ticket_id].send_event('task.progress.%s' % ticket_id, data)
 
+    def task_kill(self, ticket_id):
+
+        if ticket_id in self.tasks_running:
+            log.msg('Taks is running - killing')
+            self.tasks_running[ticket_id]['defered'].cancel()
+            return True
+        else:
+            log.msg('Taks not running - not killing')
+            return False
+
+    def task_list(self):
+        return [{
+                    'id': task_id,
+                    'name': task['name'],
+                    'args': task['args'],
+                    'kwargs': task['kwargs'],
+                } for task_id, task in self.tasks_running.items()]
+
+    def kill_client_tasks(self, client):
+        for ticket_id, task_client in self.ticket_map.items():
+            if task_client == client:
+                self.task_kill(ticket_id)
+
+
     @inlineCallbacks
     def task_start(self, client, task_name, *args, **kwargs):
         """
@@ -72,6 +105,13 @@ class ApiRpcServer(object):
                 task_defered.addCallback(self.task_completed, ticket_id)
                 task_defered.addErrback(self.task_failed, ticket_id)
 
+                self.tasks_running[ticket_id] = {
+                    'defered': task_defered,
+                    'name': task_name,
+                    'args': args,
+                    'kwargs': kwargs,
+                }
+
             except Exception as e:
                 self.task_failed(e.message, ticket_id)
 
@@ -95,7 +135,6 @@ class ApiRpcServer(object):
         return d
 
 
-
 class MdcloudWebsocketServerProtocol(WebSocketServerProtocol):
     def __init__(self):
         pass
@@ -110,9 +149,8 @@ class MdcloudWebsocketServerProtocol(WebSocketServerProtocol):
         self.factory.server.on_client_disconnect(self, wasClean, code, reason)
 
     def onMessage(self, payload, isBinary):
-
         log.msg('Websocket server in: %s' % payload)
-        #print(self.factory.server.on_message())
+        # print(self.factory.server.on_message())
 
         reactor.callLater(0, self.factory.server.on_message, self, payload, isBinary)
 
@@ -135,14 +173,13 @@ class Server(object):
     settings = inject.attr('settings')
 
     def __init__(self, port=7080):
-        self.task_map = {}
         self.port = port
         self.clients = []
 
 
     @inlineCallbacks
     def on_message(self, client, payload, is_binary=False):
-        #"""
+        # """
         #Method is called when new message arrives from client
         #"""
         #ticket_id = yield self.redis.incr('mfcloud-ticket-id')
@@ -155,6 +192,13 @@ class Server(object):
             if data['task'] == 'ping':
                 yield client.send_response(data['id'], 'pong')
 
+            elif data['task'] == 'kill':
+                success = self.rpc_server.task_kill(int(data['kwargs']['ticket_id']))
+                yield client.send_response(data['id'], success)
+
+            elif data['task'] == 'list':
+                yield client.send_response(data['id'], self.rpc_server.task_list())
+
             elif data['task'] == 'task_start':
                 ticket_id = yield self.rpc_server.task_start(client, *data['args'], **data['kwargs'])
                 yield client.send_response(data['id'], ticket_id)
@@ -165,7 +209,6 @@ class Server(object):
             log.err()
 
 
-
     def on_client_connect(self, client):
         """
         Method is called when new client is here
@@ -180,6 +223,8 @@ class Server(object):
         if client in self.clients:
             self.clients.remove(client)
 
+        self.rpc_server.kill_client_tasks(client)
+
         log.msg('Client disconnected')
 
     def shutdown(self):
@@ -219,11 +264,12 @@ class Server(object):
 
         try:
 
-            if self.settings.ssl.enabled:
+            if self.settings and self.settings.ssl.enabled:
 
                 from OpenSSL import SSL
 
                 from twisted.internet import ssl
+
                 myContextFactory = ssl.DefaultOpenSSLContextFactory(
                     self.settings.ssl.key, self.settings.ssl.cert
                 )
@@ -242,9 +288,6 @@ class Server(object):
             log.err()
 
 
-
-
-
 class MdcloudWebsocketClientProtocol(WebSocketClientProtocol):
     def __init__(self):
         pass
@@ -276,7 +319,6 @@ class MdcloudWebsocketClientProtocol(WebSocketClientProtocol):
 
 
 class Client(object):
-
     def __init__(self, host='127.0.0.1', port=7080, settings=None):
         self.port = port
         self.host = host
@@ -328,8 +370,11 @@ class Client(object):
 
                 if task_id in self.task_map:
                     method = 'on_%s' % etype
-                    # call one of on_progress, on_failure, on_success
-                    getattr(self.task_map[task_id], method)(data['data'])
+                    try:
+                        # call one of on_progress, on_failure, on_success
+                        getattr(self.task_map[task_id], method)(data['data'])
+                    except AlreadyCalledError:
+                        log.msg('Callback alredy called: %s. Skipping' % method)
                 else:
                     raise Exception('Unknown task id: %s' % task_id)
 
@@ -341,8 +386,9 @@ class Client(object):
 
         self.onc = defer.Deferred()
 
-        if self.settings.ssl.enabled:
+        if self.settings and self.settings.ssl.enabled:
             from mfcloud.ssl import CtxFactory
+
             reactor.connectSSL(self.host, self.port, factory, CtxFactory())
         else:
             reactor.connectTCP(self.host, self.port, factory)
@@ -373,8 +419,19 @@ class Client(object):
         if not wasClean:
             print('Connection closed: %s (code: %s)' % (reason, code))
 
-        # reactor.stop()
+            # reactor.stop()
 
+    @inlineCallbacks
+    def terminate_task(self, task_id):
+        result = yield self.call_sync('kill', ticket_id=task_id)
+
+        if result and task_id in self.task_map:
+            self.task_map[task_id].is_running = False
+
+        defer.returnValue(result)
+
+    def task_list(self):
+        return self.call_sync('list')
 
     @inlineCallbacks
     def call(self, task, *args, **kwargs):
@@ -394,13 +451,11 @@ class Client(object):
         defer.returnValue(task)
 
 
-
 class TaskFailure(Exception):
     pass
 
 
 class Task(object):
-
     def __init__(self, name):
         super(Task, self).__init__()
 
@@ -434,6 +489,7 @@ class Task(object):
         self.wait = defer.Deferred()
         return self.wait
 
+
 if __name__ == '__main__':
     from twisted.python import log
 
