@@ -11,6 +11,9 @@ from mfcloud.util import Interface
 import re
 from twisted.internet import defer
 from twisted.internet.defer import inlineCallbacks
+from twisted.internet.protocol import Protocol
+from twisted.web._newclient import ResponseDone
+from twisted.web.http import PotentialDataLoss
 
 logger = logging.getLogger('mfcloud.docker')
 
@@ -141,10 +144,10 @@ class DockerTwistedClient(object):
         return d
 
 
-    def logs(self, container_id, on_log):
+    def logs(self, container_id, on_log, tail=0, follow=True):
         r = self._get('containers/%s/logs' % bytes(container_id), response_handler=None, data={
-            'follow': True,
-            'tail': 0,
+            'follow': follow,
+            'tail': tail,
             'timestamps': 0,
             'stdout': True,
             'stderr': True
@@ -161,6 +164,58 @@ class DockerTwistedClient(object):
         r.addBoth(on_result)
         return r
 
+    def attach(self, container_id):
+
+
+        class Attach(Protocol):
+
+            def __init__(self, finished):
+                self.finished = finished
+
+            def dataReceived(self, data):
+                print data
+                print self.transport
+                # self.transport.write('a')
+
+
+            def connectionLost(self, reason):
+                print reason
+                if reason.check(ResponseDone):
+                    self.finished.callback(None)
+                elif reason.check(PotentialDataLoss):
+                    # http://twistedmatrix.com/trac/ticket/4840
+                    self.finished.callback(None)
+                else:
+                    self.finished.errback(reason)
+
+        r = self._post('containers/%s/attach' % bytes(container_id), response_handler=None, data={
+            # 'logs': 1,
+            'stream': 1,
+            'stdout': 1,
+            # 'stderr': 1,
+            'stdin': 1
+        })
+
+        def on_result(result):
+
+
+            if result.code == 200:
+                print('collect')
+                d = defer.Deferred()
+                protocol = Attach(d)
+                result.deliverBody(protocol)
+                return d
+                # def on_log(dat):
+                #     print dat
+                # return txhttp.collect(result, on_log)
+            elif result.code == 404:
+                return self.collect_to_exception(NotFound, result)
+            else:
+                return self.collect_to_exception(CommandFailed, result)
+
+        r.addBoth(on_result)
+        return r
+
     def events(self, on_event):
         r = self._get('events', response_handler=None)
         r.addCallback(txhttp.collect, on_event)
